#!/usr/bin/perl
###############################################################################
#
# Copyright (C) 2013-2017 Cisco and/or its affiliates. All rights reserved.
#
# THE PRODUCT AND DOCUMENTATION ARE PROVIDED AS IS WITHOUT WARRANTY
# OF ANY KIND, AND CISCO DISCLAIMS ALL WARRANTIES AND REPRESENTATIONS,
# EXPRESS OR IMPLIED, WITH RESPECT TO THE PRODUCT, DOCUMENTATION AND
# RELATED MATERIALS INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE; WARRANTIES
# ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE; AND WARRANTIES
# CONCERNING THE NON-INFRINGEMENT OF THIRD PARTY RIGHTS.
#
# IN NO EVENT SHALL CISCO BE LIABLE FOR ANY DAMAGES RESULTING FROM
# LOSS OF DATA, LOST PROFITS, LOSS OF USE OF EQUIPMENT OR LOST CONTRACTS
# OR FOR ANY SPECIAL, INDIRECT, INCIDENTAL, PUNITIVE, EXEMPLARY OR
# CONSEQUENTIAL DAMAGES IN ANY WAY ARISING OUT OF OR IN CONNECTION WITH
# THE USE OR PERFORMANCE OF THE PRODUCT OR DOCUMENTATION OR RELATING TO
# THIS AGREEMENT, HOWEVER CAUSED, EVEN IF IT HAS BEEN MADE AWARE OF THE
# POSSIBILITY OF SUCH DAMAGES.  CISCO'S ENTIRE LIABILITY TO LICENSEE,
# REGARDLESS OF THE FORM OF ANY CLAIM OR ACTION OR THEORY OF LIABILITY
# (INCLUDING CONTRACT, TORT, OR WARRANTY), SHALL BE LIMITED TO THE
# LICENSE FEES PAID BY LICENSEE TO USE THE PRODUCT.
#
###############################################################################
#
#  Change Log
#
#  23-June-2016  - Adam Hogan (unsure of original date. It is older.)
#    o Initial Release
#  09-March-2017 - Mark Cairns
#    o Update Copyright
#    o Adjust remarks for file locations
#    o Add debugging log functions (commented out by default)
#    o Alter exit status codes to match 6.x API - GUI Correlation Status Results
#    o Add custom exit codes to module.template - GUI Correlation Status Results
#    o Change logic to return custom status results
#    o Add ability to restrict and define host limit for local blacklist in
#        module.template. Can be configured per instance, preventing
#        infinite file size. Old entries are pruned as FIFO. Instances can be 
#        modified at a later time to reduce or increase host limits
#
#
#
#
#
###############################################################################
#
# Blacklist Remediation module for Firepower Management Center.
#
# Adam Hogan             - Unknown Date - Version 1.0
# Mark Cairns            - 03/09/2017 - Version 1.1
#
#
#      Creates a custom remediation action allowing a correlation
#      rule to identify an event, pull the IP (source or destination,
#      user's choice), and add it to a custom blacklist.
#
#      See README for info on using this module
#
###############################################################################

# Comment lines with "LogThis" function are for debugging to log.txt file
# Uncomment as needed. Verify my $log = 1 below.
# View log.txt file located in the following directory:
# /var/sf/remediations/yourmodulename_version/yourinstancename/

# Configuring a remediation from module.template will generate a section in:
# /etc/sf/remediation/remediation.conf
# Custom exit status values are loaded and can be viewed in this file

# Local blacklist files for security intel will be generated in /var/sf/htdocs

# Custom Security Intelligence feeds can be checked no faster than every
# 30 minutes. New IPs added to the list may take 30 minutes or longer before
# showing up in the SI feed, depending on settings and time since last query.

# Use XML::Smart to access the XML data.
use XML::Smart;
use Tie::File qw();
use File::Glob;
use strict;
use warnings;

# Error codes defined in the API
use constant INPUT_ERR=>1;
use constant CONFIG_ERR=>2;
use constant UNREACHABLE=>7;
use constant FAILED_ERR=>11;
use constant FAILED_MD5=>133;

# Custom Error codes defined in version 1.1 module.template
# Verify in remediation.conf as listed above
# 130 Local IP Blacklist Successful
# 131 Remote IP Blacklist Successful
# 132 IP Already Blacklisted - Operation Aborted
# 133 MD5 Hash Calculation Error
# 134 Local IP Blacklist Successful - host restriction limit - pruning SI list


# Should we log?
my $log = 0;
my $debug = 0;

if ($debug) {
  use Data::Dumper;
}

# GetXMLConfig
# Extract instance and remediations configuration from instance.conf
sub GetXMLConfig($)
{
    my $rem_name = shift;
    my $valid_rem=0;
    my %inst_config;
    my %rem_config;

    my $XML = XML::Smart->new("./instance.conf");

    # Get the instance configuration
    my @config = $XML->{instance}->{config}->nodes();
    foreach my $config (@config)
    {
        $inst_config{$config->{name}} = $config->content;
    }
    undef @config;

    # Get the configuration for the named remediation
    foreach my $remediation (@{$XML->{instance}->{remediation}})
    {
        #if($remediation->{type} =~ /^$rem_name$/)
        if($remediation->{name} =~ /^$rem_name$/)
        {
            $valid_rem=1;
            my @config = $remediation->{config}->nodes();
            foreach my $config (@config)
            {
                $rem_config{$config->{name}} = $config->content;
            }

            $rem_config{'type'}=$remediation->{type}->content;
            last;
        }
    }
    if ($valid_rem == 0)
    {
        warn "Invalid remediation type.";
        exit(CONFIG_ERR);
    }
    return(\%inst_config,\%rem_config);
}

###############################################################################
## Check to see if the IP in question is already in that list.
## Is this important? Maybe not. Cleaner though...
###############################################################################

sub IPAlreadyExists {
    ## should this work with a filehandle instead?
    my $ip = shift;
    my $file = shift;

    my $found = 0;

    open (LIST,'<',$file);
    while(<LIST>){
        next unless $_ =~ /$ip(\n)?$/;
        $found++;
    }

    return $found;
}
###############################################################################
## End IPAlready Exists
###############################################################################

###############################################################################
#
# Add the IP passed to a local list
#
# This will add the IP to a feed which can be loaded from URL
# https://localhost/<file>
#
###############################################################################

sub BlacklistLocal
{
    my $instance_config = shift;
    my $remediation_config = shift;
    my $pe_event_data = shift;
    my $offender = "";
    my $filelength = 0;
    my $fileoverage = 0;

    # Pull IP address from event data.
    ($offender) = @$pe_event_data;
    if (!defined($offender)) {
        warn "No IP specified!\n";
        &help;
        exit(CONFIG_ERR);
    }

    # Get filenames from the config
    my $file = $remediation_config->{FileName};
    my $hashfile = $remediation_config->{md5FileName};
        my $filerestrict = $remediation_config->{FileRestrict};
        my $filelimit = $remediation_config->{FileLimit};

    my $path = "/var/sf/htdocs/";
    my $fq = $path . $file;
    my $fqhash = $path . $hashfile;

    ## Logging
    if ($log) {
       LogThis("BlacklistLocal called to block $offender in $fq");
    }

    if (IPAlreadyExists($offender, $fq)) {
            return 132;
                }
        else {
       #open file
       ##TODO path/name validation
       open(LIST, ">>$fq");

       #append $offender to file
       print LIST "$offender\n";

       $filelength = ($. + 1);

       # close file
       close(LIST);

       if ($filerestrict) {   # did you set a file limit in the instance configuration?

           if ($filelength >= $filelimit){    # if true, rip out lines to maintain file size
                   $fileoverage= ($filelength - $filelimit);

                   for my $filename (glob $fq) {
                   tie my @file, 'Tie::File', $filename or die "Could not open $filename: $!";
                   splice (@file, 0, $fileoverage);
                }
				   #generate MD5 file
                   system("md5sum $fq > $fqhash");
                   if ($? == -1) {
                   print "system md5 failed: $!";
                   exit(FAILED_MD5);
				}
				   #return custom success code to main script and pass to exit value $rval
				
				   return 134;
				}
            }
        }

       #generate MD5 file
       system("md5sum $fq > $fqhash");
       if ($? == -1) {
           print "system md5 failed: $!";
           exit(FAILED_MD5);
       }

    #return custom success code to main script and pass to exit value $rval
	
	return 130;
}
###############################################################################
# End BlacklistLocal
###############################################################################

###############################################################################
# BlacklistRemote - Add the IP passed to a remote list
###############################################################################
sub BlacklistRemote {

    my $instance_config = shift;
    my $remediation_config = shift;
    my $pe_event_data = shift;
    my $offender = "";

    # Pull IP address from event data.
    ($offender) = @$pe_event_data;
    if (!defined($offender)) {
        warn "No IP specified!\n";
        &help;
        exit(CONFIG_ERR);
    }

    # Get date from the config
    # host, user, file, md5 file
    my $file = $remediation_config->{FileName};
    my $hashfile = $remediation_config->{md5FileName};
    my $host = $remediation_config->{HostName};
    my $user = $remediation_config->{UserName};

    ## Logging
    if ($log) {
       LogThis("BlacklistRemote called to block: $offender in $file on $host");
    }

    my $cmd = "ssh $user" . '@' . "$host";
    $cmd .= " 'echo $offender >> $file ";
    $cmd .= " && md5sum $file > $hashfile '";

    if ($log) {
       LogThis("cmd to execute",$cmd);
    }

    my $out = `$cmd`;
    ### TODO output and error from ssh. Permissions?

    if ($log) {
       LogThis("cmd output", $out);
    }

    #return custom success code to main script and pass to exit value $rval
    return 131;
}
###############################################################################
## End BlacklistRemote
###############################################################################

# Logging Routing
sub LogThis {

   open(LOG, ">>log.txt");

   foreach (@_) {
      print LOG $_ . "\t";
   }

   print LOG "\n";

   close(LOG);
}


# Print a usage message
sub help
{
    warn "usage: blacklist.pl <Instance Name> [policy event data...]\nThough this is usually called by the DC's
rememdiation mod.";
    exit(INPUT_ERR);
}

###############################################################################
# Starting the Main function
###############################################################################

# Simple usage check.
if(@ARGV < 2)
{
    &help;
}

# Log the program's call
if ($log) {
    my $now = localtime;
    LogThis($now, "program called", $ARGV[0], $ARGV[1]);
}

# Getting the Command-line arguments
# These should be the type of the module followed by all event data passed
# according to the module.config

my $rem_name = shift;  # ARGV[0]
my @pe_event_data = @ARGV;

my ($instance_config, $remediation_config) = &GetXMLConfig($rem_name);

if((! defined($instance_config))||(! defined($remediation_config)))
{
    warn "Problems reading the instance configuration\n";
    exit(CONFIG_ERR);
}

if ($debug) {
  print "instance config:\n";
  print Dumper $instance_config;
  print "remediation config";
  print Dumper $remediation_config;
}

# Call the appropriate function for the remediation type and set initial $rval
# for exit status to 8. This is "The remediation was attempted but failed."
# $rval will be rewritten with a successful return code upon completion of a
# remediation sub-routine.

my $rval = 8;  # Sys defined exit value - The remediation was attempted but failed.
if($remediation_config->{type} eq "local_src_blacklist")
{
# LogThis("rval is set to $rval before calling Source IP BlacklistLocal Sub");

    $rval = BlacklistLocal($instance_config, $remediation_config,
            \@pe_event_data);

# LogThis("Source IP BlacklistLocal Sub completed and returned rval of $rval");
}
elsif($remediation_config->{type} eq "local_dst_blacklist")
{
# LogThis("rval is set to $rval before calling Dest IP BlacklistLocal Sub");

    $rval = BlacklistLocal($instance_config, $remediation_config,
            \@pe_event_data);

# LogThis("Dest IP BlacklistLocal Sub completed and returned rval of $rval");
}
elsif($remediation_config->{type} eq "remote_src_blacklist")
{
# LogThis("rval is set to $rval before calling Source IP BlacklistRemote Sub");

    $rval = BlacklistRemote($instance_config, $remediation_config,
            \@pe_event_data);

# LogThis("Source IP BlacklistRemote completed and returned rval of $rval");
}
elsif($remediation_config->{type} eq "remote_dst_blacklist")
{
# LogThis("rval is set to $rval before calling Dest IP BlacklistRemote Sub");

    $rval = BlacklistRemote($instance_config, $remediation_config,
            \@pe_event_data);

# LogThis("Dest IP BlacklistRemote completed and returned rval of $rval");
}
else
{
    my $t = $remediation_config->{type};
    warn "Invalid remediation type ($t). Check your instance.conf\n";
    exit(CONFIG_ERR);
}
# LogThis("The exit code being returned to the remediation module is $rval");

exit($rval);
